package betterwithmods.common.blocks.mechanical.tile;

import betterwithmods.api.block.ITurnable;
import betterwithmods.api.capabilities.CapabilityMechanicalPower;
import betterwithmods.api.tile.IMechanicalPower;
import betterwithmods.common.BWMBlocks;
import betterwithmods.common.BWMRecipes;
import betterwithmods.common.blocks.tile.IMechSubtype;
import betterwithmods.common.blocks.tile.TileBasic;
import betterwithmods.common.registry.blockmeta.managers.TurntableManager;
import betterwithmods.common.registry.blockmeta.recipe.TurntableRecipe;
import betterwithmods.util.DirUtils;
import betterwithmods.util.InvUtils;
import betterwithmods.util.MechanicalUtil;
import net.minecraft.block.*;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.init.Items;
import net.minecraft.init.SoundEvents;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.util.*;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.text.TextComponentTranslation;
import net.minecraft.world.World;
import net.minecraft.world.WorldServer;
import net.minecraftforge.common.capabilities.Capability;

import javax.annotation.Nonnull;

public class TileEntityTurntable extends TileBasic implements IMechSubtype, ITickable, IMechanicalPower {
    private static final int[] ticksToRotate = {10, 20, 40, 80};
    public byte timerPos = 0;
    private int potteryRotation = 0;
    private boolean potteryRotated = false;
    private double[] offsets = {0.25D, 0.375D, 0.5D, 0.625D};
    private boolean asynchronous = false;
    private int rotationTime = 0;
    private int power;

    @Override
    public void func_145839_a(NBTTagCompound tag) {
        super.func_145839_a(tag);
        if (tag.func_74764_b("SwitchSetting")) {
            this.timerPos = tag.func_74771_c("SwitchSetting");
            if (this.timerPos > 3)
                this.timerPos = 3;
        }
        if (tag.func_74764_b("PotteryRotation"))
            this.potteryRotation = tag.func_74762_e("PotteryRotation");
        if (tag.func_74764_b("Asynchronous"))
            this.asynchronous = tag.func_74767_n("Asynchronous");
        if (tag.func_74764_b("RotationTime"))
            this.rotationTime = tag.func_74762_e("RotationTime");
        this.power = tag.func_74762_e("power");
    }

    @Override
    public NBTTagCompound func_189515_b(NBTTagCompound tag) {
        super.func_189515_b(tag);
        tag.func_74768_a("PotteryRotation", this.potteryRotation);
        tag.func_74774_a("SwitchSetting", this.timerPos);
        tag.func_74757_a("Asynchronous", this.asynchronous);
        if (this.asynchronous || this.rotationTime != 0)
            tag.func_74768_a("RotationTime", this.rotationTime);
        tag.func_74768_a("power", power);
        return tag;
    }

    @Override
    public boolean shouldRefresh(World world, BlockPos pos, IBlockState oldState, IBlockState newState) {
        return oldState.func_177230_c() != newState.func_177230_c();
    }

    @Override
    public void func_73660_a() {
        if (!this.func_145831_w().field_72995_K) {
            this.power = calculateInput();
            if (power > 0) {
                if (!asynchronous && func_145831_w().func_82737_E() % (long) ticksToRotate[timerPos] == 0) {
                    this.func_145831_w().func_184133_a(null, field_174879_c, SoundEvents.field_187885_gS, SoundCategory.BLOCKS, 0.05F, 1.0F);
                    rotateTurntable();
                } else if (asynchronous) {
                    rotationTime++;
                    if (rotationTime >= ticksToRotate[timerPos]) {
                        rotationTime = 0;
                        this.func_145831_w().func_184133_a(null, field_174879_c, SoundEvents.field_187885_gS, SoundCategory.BLOCKS, 0.05F, 1.0F);
                        rotateTurntable();
                    }
                }
            }
        }
    }

    public boolean processRightClick(EntityPlayer player) {
        if (!player.func_184586_b(EnumHand.MAIN_HAND).func_190926_b()) {
            if (player.func_184586_b(EnumHand.MAIN_HAND).func_77973_b() == Items.field_151113_aN) {
                toggleAsynchronous(player);
                return true;
            }
        } else if (player.func_184614_ca().func_190926_b()) {
            advanceTimerPos();
            func_145831_w().func_180497_b(field_174879_c, this.func_145838_q(), this.func_145838_q().func_149738_a(func_145831_w()), 5);
            func_145831_w().func_184133_a(null, field_174879_c, SoundEvents.field_187885_gS, SoundCategory.BLOCKS, 0.3F, 0.6F);
            return true;
        }
        return false;
    }

    public void toggleAsynchronous(EntityPlayer player) {
        if (!this.func_145831_w().func_82736_K().func_82766_b("doDaylightCycle")) {
            if (!asynchronous) {
                this.asynchronous = true;
            } else if (player != null) {
                player.func_146105_b(new TextComponentTranslation("message.bwm:async.unavailable"), false);
            }
        } else {
            boolean isSneaking = player.func_70093_af();
            String isOn = "enabled";
            boolean async = !this.asynchronous;
            if ((!async && !isSneaking) || (async && isSneaking))
                isOn = "disabled";
            player.func_146105_b(new TextComponentTranslation("message.bwm:async." + isOn), false);
            if (!isSneaking) {
                this.func_145831_w().func_184133_a(null, field_174879_c, SoundEvents.field_187839_fV, SoundCategory.BLOCKS, 0.05F, 1.0F);
                this.asynchronous = async;
            }
        }
    }

    public void rotateTurntable() {
        boolean reverse = MechanicalUtil.isRedstonePowered(field_145850_b,field_174879_c);

        this.potteryRotated = false;

        for (int tempY = 1; tempY < 3; tempY++) {
            BlockPos searchPos = field_174879_c.func_177982_a(0, tempY, 0);
            boolean canTransmitHorizontally = canBlockTransmitRotationHorizontally(searchPos);
            boolean canTransmitVertically = canBlockTransmitRotationVertically(searchPos);
            rotateBlock(searchPos, reverse);

            if (canTransmitHorizontally)
                rotateBlocksAttached(searchPos, reverse);

            if (!canTransmitVertically)
                break;
        }

        if (!potteryRotated)
            potteryRotation = 0;

        func_145831_w().func_190524_a(field_174879_c, BWMBlocks.SINGLE_MACHINES, field_174879_c);
    }

    public byte getTimerPos() {
        return timerPos;
    }

    public double getOffset() {
        return offsets[this.timerPos];
    }

    public void advanceTimerPos() {
        timerPos++;
        if (timerPos > 3)
            timerPos = 0;
        IBlockState state = func_145831_w().func_180495_p(field_174879_c);
        func_145831_w().func_184138_a(field_174879_c, state, state, 3);
    }

    private boolean canBlockTransmitRotationHorizontally(BlockPos pos) {
        Block target = func_145831_w().func_180495_p(pos).func_177230_c();

        if (target instanceof ITurnable) {
            return ((ITurnable) target).canRotateHorizontally(func_145831_w(), pos);
        }
        if (target == Blocks.field_150359_w || target == Blocks.field_150399_cn)
            return true;
        if (target instanceof BlockPistonBase) {
            IBlockState state = func_145831_w().func_180495_p(pos);

            return !state.func_177229_b(BlockPistonBase.field_176320_b);
        }
        return !(target == Blocks.field_180384_M || target == Blocks.field_150332_K) && this.func_145831_w().func_175677_d(pos, false);

    }

    private boolean canBlockTransmitRotationVertically(BlockPos pos) {
        Block target = func_145831_w().func_180495_p(pos).func_177230_c();

        if (target instanceof ITurnable) {
            return ((ITurnable) target).canRotateVertically(func_145831_w(), pos);
        }
        if (target == Blocks.field_150359_w)
            return true;
        if (target == Blocks.field_150435_aG)
            return false;
        if (target instanceof BlockPistonBase) {
            IBlockState state = func_145831_w().func_180495_p(pos);

            return !state.func_177229_b(BlockPistonBase.field_176320_b);
        }
        return !(target == Blocks.field_180384_M || target == Blocks.field_150332_K) && this.func_145831_w().func_175677_d(pos, false);

    }

    private void rotateBlocksAttached(BlockPos pos, boolean reverse) {
        Block[] newBlocks = new Block[4];
        for (int i = 0; i < 4; i++)
            newBlocks[i] = null;

        for (int i = 2; i < 6; i++) {
            BlockPos offset = pos.func_177972_a(EnumFacing.func_82600_a(i));

            Block block = func_145831_w().func_180495_p(offset).func_177230_c();
            IBlockState state = func_145831_w().func_180495_p(offset);
            boolean attached = false;

            if (block instanceof BlockTorch) {
                EnumFacing facing = state.func_177229_b(BlockTorch.field_176596_a);
                if ((facing == EnumFacing.func_82600_a(i))) {
                    attached = true;

                    if (block == Blocks.field_150437_az)
                        block = Blocks.field_150429_aA;
                }
            } else if (block instanceof BlockLadder) {
                int meta = state.func_177229_b(BlockLadder.field_176382_a).func_176745_a();
                if (meta == i)
                    attached = true;
            } else if (block == Blocks.field_150444_as) {
                int meta = state.func_177229_b(BlockWallSign.field_176412_a).func_176745_a();
                if (meta == i) {
                    block.func_176226_b(func_145831_w(), offset, state, 0);
                    this.func_145831_w().func_175698_g(offset);
                }
            } else if (block instanceof BlockButton) {
                EnumFacing facing = state.func_177229_b(BlockButton.field_176387_N);
                if ((facing == EnumFacing.func_82600_a(i))) {
                    block.func_176226_b(func_145831_w(), offset, state, 0);
                    this.func_145831_w().func_175698_g(offset);
                }
            } else if (block == Blocks.field_150442_at) {
                BlockLever.EnumOrientation facing = state.func_177229_b(BlockLever.field_176360_a);
                if (facing.func_176852_c() == EnumFacing.func_82600_a(i)) {
                    block.func_176226_b(func_145831_w(), offset, state, 0);
                    this.func_145831_w().func_175698_g(offset);
                }
            }

            if (attached) {
                EnumFacing destFacing = DirUtils.rotateFacingAroundY(EnumFacing.func_82600_a(i), reverse);
                newBlocks[destFacing.ordinal() - 2] = block;
                func_145831_w().func_175698_g(offset);
            }
        }

        for (int i = 0; i < 4; i++) {
            Block block = newBlocks[i];

            if (block != null) {
                int facing = i + 2;
                int meta;

                BlockPos offset = pos.func_177972_a(EnumFacing.func_82600_a(facing));
                IBlockState state = func_145831_w().func_180495_p(offset);
                if (block instanceof BlockTorch) {
                    int targetFacing = 0;

                    if (facing == 2)
                        targetFacing = 4;
                    else if (facing == 3)
                        targetFacing = 3;
                    else if (facing == 4)
                        targetFacing = 2;
                    else if (facing == 5)
                        targetFacing = 1;
                    meta = targetFacing;
                    state = ((BlockTorch) block).func_176203_a(meta);
                } else if (block instanceof BlockLadder) {
                    meta = facing;
                    state = ((BlockLadder) block).func_176203_a(meta);
                }

                if (func_145831_w().func_180495_p(offset).func_177230_c().func_176200_f(func_145831_w(), offset))
                    func_145831_w().func_175656_a(offset, state);
                else {
                    EnumFacing oldFacing = DirUtils.rotateFacingAroundY(EnumFacing.func_82600_a(facing), !reverse);
                    BlockPos oldPos = pos.func_177972_a(oldFacing);
                    block.func_176226_b(func_145831_w(), oldPos, state, 0);
                }
            }
        }
    }

    private void rotateBlock(BlockPos pos, boolean reverse) {
        if (func_145831_w().func_175623_d(pos))
            return;

        IBlockState state = func_145831_w().func_180495_p(pos);
        ItemStack stack = new ItemStack(state.func_177230_c(), 1, state.func_177230_c().func_180651_a(state));
        Block target = state.func_177230_c();
        Rotation rot = reverse ? Rotation.COUNTERCLOCKWISE_90 : Rotation.CLOCKWISE_90;

        if (TurntableManager.INSTANCE.contains(stack) && TurntableManager.INSTANCE.getRecipe(stack) != null) {
            if (target instanceof ITurnable) {
                ITurnable block = (ITurnable) target;
                if (block.canRotateOnTurntable(func_145831_w(), pos))
                    block.rotateAroundYAxis(func_145831_w(), pos, reverse);
            } else if (state != state.func_185907_a(rot))
                func_145831_w().func_175656_a(pos, state.func_185907_a(rot));
            rotateCraftable(state, TurntableManager.INSTANCE.getRecipe(stack), pos, reverse);
            this.potteryRotated = true;
        } else if (target instanceof ITurnable) {
            ITurnable block = (ITurnable) target;

            if (block.canRotateOnTurntable(func_145831_w(), pos))
                block.rotateAroundYAxis(func_145831_w(), pos, reverse);
        } else {
            IBlockState newState = state.func_185907_a(rot);
            if (!(target instanceof BlockRailBase) && !(target instanceof BlockPistonExtension) && !(target instanceof BlockPistonMoving) && state != newState) {
                if (target instanceof BlockPistonBase) {
                    if (!state.func_177229_b(BlockPistonBase.field_176320_b))
                        func_145831_w().func_175656_a(pos, newState);
                } else
                    func_145831_w().func_175656_a(pos, newState);
            } else if (target instanceof BlockRailBase) {
                BlockRailBase rail = (BlockRailBase) target;
                BlockRailBase.EnumRailDirection dir = state.func_177229_b(rail.func_176560_l());
                if (dir != BlockRailBase.EnumRailDirection.ASCENDING_NORTH && dir != BlockRailBase.EnumRailDirection.ASCENDING_EAST && dir != BlockRailBase.EnumRailDirection.ASCENDING_SOUTH && dir != BlockRailBase.EnumRailDirection.ASCENDING_WEST) {
                    if (state != newState)
                        func_145831_w().func_175656_a(pos, newState);
                }
            }
        }
    }

    private void spawnParticles(IBlockState state) {
        ((WorldServer)this.field_145850_b).func_175739_a(EnumParticleTypes.BLOCK_DUST, field_174879_c.func_177958_n()+0.5, field_174879_c.func_177956_o()+1, field_174879_c.func_177952_p()+0.5,30, 0.0D, 0.5D, 0.0D, 0.15000000596046448D, Block.func_176210_f(state));
    }

    private void rotateCraftable(IBlockState input, TurntableRecipe craft, BlockPos pos, boolean reverse) {
        Block block = input.func_177230_c();
        this.potteryRotation++;
        if (this.potteryRotation > 7) {
            if (!craft.getOutputs().isEmpty() && craft.getOutputs().size() > 0) {
                for (ItemStack scrap : craft.getOutputs()) {
                    InvUtils.ejectStackWithOffset(func_145831_w(), pos.func_177984_a(), scrap.func_77946_l());
                }
            }
            func_145831_w().func_175656_a(pos, BWMRecipes.getStateFromStack(craft.getResult()));
            this.potteryRotation = 0;

        }
        spawnParticles(input);
        this.func_145831_w().func_184133_a(null, pos, block.getSoundType(input, this.func_145831_w(), pos, null).func_185841_e(), SoundCategory.BLOCKS, 0.5F, func_145831_w().field_73012_v.nextFloat() * 0.1F + 0.8F);
    }

    @Override
    public int getSubtype() {
        return this.timerPos + 8;
    }

    @Override
    public void setSubtype(int type) {
        this.timerPos = (byte) Math.min(type, 3);
    }


    @Override
    public int getMechanicalOutput(EnumFacing facing) {
        return -1;
    }

    @Override
    public int getMechanicalInput(EnumFacing facing) {
        if (facing == EnumFacing.DOWN)
            return MechanicalUtil.getPowerOutput(field_145850_b, field_174879_c.func_177972_a(facing), facing.func_176734_d());
        return 0;
    }

    @Override
    public int getMaximumInput(EnumFacing facing) {
        return 1;
    }

    @Override
    public int getMinimumInput(EnumFacing facing) {
        return 0;
    }

    @Override
    public boolean hasCapability(@Nonnull Capability<?> capability, @Nonnull EnumFacing facing) {
        if (capability == CapabilityMechanicalPower.MECHANICAL_POWER)
            return true;
        return super.hasCapability(capability, facing);
    }

    @Nonnull
    @Override
    public <T> T getCapability(@Nonnull Capability<T> capability, @Nonnull EnumFacing facing) {
        if (capability == CapabilityMechanicalPower.MECHANICAL_POWER)
            return CapabilityMechanicalPower.MECHANICAL_POWER.cast(this);
        return super.getCapability(capability, facing);
    }
}
